/*
  GPS Mini Navigator    for navigating predefined ways & extended gps logging

  Copyright (C) 2008  Matthias Larisch

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License along with
  this program; if not, see <http://www.gnu.org/licenses/>.
  

  file description: simple filesystem implementation

*/


#include "egl.h"
#include "filesystem.h"
#include "mmc.h"
//#include "suart.h"


/* This filesystem library is about 1,5 kByte in compiled size. It needs modified mmc-library with
   part-sector-read.
   features:
   -Up to 255 files
   -Opening / reading files (up to 255 bytes per read call)
   -Creating / writing files (up to 512 bytes per write call)
   -read possible while writing but not the other way round
   -2 files can be opened for writing but software has to call a changer which safes buffer to card
    (changing should not be done that often else a short card lifetime will be expected)
   -no file end markers: read after "EOF" will return data until sector end
*/

struct
{
  uint8_t number;
  union
  {
    uint16_t w;
	uint8_t b[2];
  } cluster;
  uint8_t sector_of_cluster;
  uint16_t remaining_bytes_in_sector;
} current_file;

struct write_file_t
{
  union
  {
    uint16_t w;
	uint8_t b[2];
  } cluster;
  uint8_t sector_of_cluster;
  uint8_t *bufferw;
};

union buffer_t buffer;
struct write_file_t write_file;
struct write_file_t paused_write_file;
//uint8_t *bufferw;
#define buffer_end buffer.b + 512

static uint16_t search_free_cluster();

// optimized assembler function to multiply an address with clustersize
// can only shift!
// output value is max. 24 bit!
static inline uint32_t addr_mul_clustersize (uint32_t cluster)
{
  asm volatile ("1:"					"\n\t"
                "ldi %D0, %2"			"\n\t"
  				"add %A0, %A0"			"\n\t"
				"adc %B0, %B0"			"\n\t"
				"adc %C0, %C0"			"\n\t"
				"dec %D0"				"\n\t"
				"brne 1b"
  				:"+r" (cluster):"0" (cluster), "M" (CLUSTER_BIT));
  return cluster;
}

// Opens a file and sets mmc pointer to first user data byte. WAYPOINT: This is first byte of log!
// Returns: file type
// DOESNT FINISH MMC TRANSFER
// Intended for read-only!
uint8_t open_file (uint8_t filenumber)
{
  uint8_t temp;
  current_file.number = filenumber;
  current_file.sector_of_cluster = 0;
  current_file.remaining_bytes_in_sector = SECTORSIZE;
  mmc_read_sector_start(filenumber);
  uint8_t retval = mmc_read_sector_continue(0,1);		// first byte = filetype
  if(retval == 0)
    return 0;
  //current_file.cluster.w = 0;    // this is not needed? we read in the word next line...
  mmc_read_sector_continue(current_file.cluster.b,2);
  if(retval == FT_WAYPOINT)
    return retval;
  mmc_read_sector_finish(0,1);
  if(current_file.cluster.w < (FILETABLESIZE / CLUSTERSIZE))	
    return 0xFF;// error in file system
    
  mmc_read_sector_start(addr_mul_clustersize(current_file.cluster.w));
  temp = mmc_read_sector_continue(0,1);
  if(temp == (retval | 0x80))		// additional error check
  {
    current_file.remaining_bytes_in_sector = SECTORSIZE - 1;
	return retval;
  }
  else
    return 0xFE;
}
    
// Creates a file for writing. (Appending not possible!)
// parameterrs: filetype
// returns: filenumber on success
// Sets current_file_position to first possible write position
// DESTROYS BUFFER (first 3 bytes), expects WAYPOINT data in buffer!
// closes opened sectors
uint8_t create_file (uint8_t filetype)
{
  write_file.cluster.w = search_free_cluster();
  uint8_t i = 0;
  uint8_t filenumber = 0;
  while (filenumber == 0)
  {
    mmc_read_sector_start(i);
    if(mmc_read_sector_continue(0,1) == 0) filenumber = i;
    mmc_read_sector_finish(0,1);
    if(i == 0xFF) return 0;		// no free file entries...
    i++;
  }
  mmc_read_sector_finish(0,1);
  buffer.b[0] = filetype;
  buffer.b[1] = write_file.cluster.b[0];
  buffer.b[2] = write_file.cluster.b[1];

  mmc_write_sector(filenumber, buffer.b);
  write_file.bufferw = &buffer.b[1];
  buffer.b[0] |= 0x80;
  write_file.sector_of_cluster = 0;
  return filenumber;
}


// writes count bytes of buffer to internal write buffer and handles all writing, sector, cluster thing.
// this is the only filesystem function that handles SD power itself
uint8_t w_file (uint8_t *Buffer, uint8_t count)
{  
  while(count--)
  {
    if((write_file.sector_of_cluster == CLUSTERSIZE - 1) && write_file.bufferw == buffer_end - 2)
	{
	  buffer.w[255] = search_free_cluster();
	  //mmc_shutdown();
	  if(buffer.w[255] == 2) return 1;		//error: medium full
	  write_file.bufferw = buffer_end;
    }
	if(write_file.bufferw == buffer_end)
	{
	  mmc_write_sector(addr_mul_clustersize(write_file.cluster.w) + write_file.sector_of_cluster++, buffer.b);
	  if(write_file.sector_of_cluster == CLUSTERSIZE)
	  {
	    write_file.sector_of_cluster = 0;
		write_file.cluster.w = buffer.w[255];
		buffer.b[0] = 0x80;
		write_file.bufferw = buffer.b+1;
      }
	  else write_file.bufferw = buffer.b;
    }      
    *write_file.bufferw++ = *Buffer++;
  }
  return 0;		// OK
}

//fills the remaining bytes of the write buffer with zeroes and writes it to file
void close_write_file()
{
  uint8_t *temp_p = write_file.bufferw;
  //for(;bufferw <= buffer_end; j++)
  //  *(bufferw+j) = 0;
  do
  {
    *temp_p++ = 0;
  } while(temp_p <= buffer_end);
  
  mmc_write_sector(addr_mul_clustersize(write_file.cluster.w) + write_file.sector_of_cluster, buffer.b);
}

// searches for free cluster, starts @ beginning of drive
// returns: cluster number of free cluster
// duration: varies. function saves last position of free cluster, so
// first run after device reset can be very long.
// you HAVE to reset device after deletion of many many files/clusters
// because cluster counter wont start from the beginning when end of drive is reached
static uint16_t search_free_cluster()
{
  static uint16_t retval = (FILETABLESIZE / CLUSTERSIZE) - 1;
  uint8_t done = 0;
  while (done == 0)
  {
    mmc_read_sector_start(addr_mul_clustersize(++retval));
    if((mmc_read_sector_continue(0,1) & 0x80) == 0) done = 1;
	mmc_read_sector_finish(0,1);
	if(retval >= CLUSTERS) retval = (FILETABLESIZE / CLUSTERSIZE);    // Medium is full
  }
  return retval;
}


// seeks forward in current file to next cluster
// returns: 1 if in next cluster
//          0 if already in last cluster
// current_file struct is always set to the beginning of a cluster
// no mmc operation will be pending
/*uint8_t seek_next_cluster()
{
  union
  {
    uint8_t b[2];
	uint16_t w;
  } temp;
  if(current_file.sector_of_cluster != CLUSTERSIZE - 1)
  {
    mmc_read_sector_finish(0,1);
	mmc_read_sector_start(current_file.cluster.w * CLUSTERSIZE + CLUSTERSIZE - 1);
	current_file.remaining_bytes_in_sector = SECTORSIZE - 2;
  }
  mmc_read_sector_continue(0, (current_file.remaining_bytes_in_sector) | 0x8000);		// discard everything to cluster address
  mmc_read_sector_finish(temp.b, 0);		// read next cluster address
  current_file.sector_of_cluster = 0;
  current_file.remaining_bytes_in_sector = 0;
  if(temp.w == 0)
    return 0;
  current_file.cluster.w = temp.w;
  return 1;
}
*/



// Seeks currently opened file offset byte forward.
// returns: 0 - ok    1 - eof reached   2 - error
// warning: NOT CLUSTER SAFE!

/*
uint8_t seek_file (uint16_t offset)
{
  if(current_file.cluster.w < FILETABLESIZE / CLUSTERSIZE)
    return 2;
  if(current_file.remaining_bytes_in_sector != 0)				// CHECK ME
  {
    if(offset >= current_file.remaining_bytes_in_sector)
	{
	  offset -= current_file.remaining_bytes_in_sector;
	  
	  current_file.sector_of_cluster++;
	  current_file.byte_of_sector = 0;
	  mmc_read_sector_finish(0,1);
	  if(offset == 0) mmc_read_sector_start(current_file.cluster.w * CLUSTERSIZE + current_file.sector_of_cluster);
    }
	else
	{
	  current_file.byte_of_sector += offset;
	  mmc_read_sector_continue(0,offset | 0x8000);
	  offset = 0;
    } 
  }
  while(offset > SECTORSIZE)
  {
    offset -= SECTORSIZE;
	current_file.sector_of_cluster++;
  }
  if(offset > 0)
  {
    mmc_read_sector_start(current_file.cluster.w * CLUSTERSIZE + current_file.sector_of_cluster);
	mmc_read_sector_continue(0,offset | 0x8000);
	current_file.byte_of_sector += offset;
  }
  return 0;
}
*/

// Reads <count> bytes from current position of file into buffer.
// Warning: Buffer needs to be 2 bytes greater than count!
// Returns: 0 on OK, 1 on eof reached with or before end of byte counter.
// 2 on error
uint8_t read_file (uint8_t *Buffer, uint8_t count)
{
  if(current_file.cluster.w < FILETABLESIZE / CLUSTERSIZE)
    return 2;
  if(count < current_file.remaining_bytes_in_sector) //-spacing))
  {
    mmc_read_sector_continue(Buffer, count);
	current_file.remaining_bytes_in_sector -= count;
	return 0;
  }
  mmc_read_sector_finish(Buffer, 0);
  Buffer += current_file.remaining_bytes_in_sector;
  current_file.sector_of_cluster++;
  if(current_file.sector_of_cluster == CLUSTERSIZE)
  {
    current_file.cluster.b[1]=*(Buffer+1);
	current_file.cluster.b[0]=*Buffer;
	current_file.sector_of_cluster = 0;
	if(current_file.cluster.w < FILETABLESIZE / CLUSTERSIZE || current_file.cluster.w > CLUSTERS)
	  return 1;
  }
  count -= current_file.remaining_bytes_in_sector;
  mmc_read_sector_start(addr_mul_clustersize(current_file.cluster.w) + current_file.sector_of_cluster);
  current_file.remaining_bytes_in_sector = 512 - count;
  if(current_file.sector_of_cluster == 0)
  {
    mmc_read_sector_continue(0,1);
	current_file.remaining_bytes_in_sector--;
  }
  uint8_t temp = mmc_read_sector_continue(Buffer,count);
  if(count == 1) *Buffer = temp;				// Workaround for "other function" with count = 1
  
  if(current_file.sector_of_cluster == CLUSTERSIZE - 1) current_file.remaining_bytes_in_sector -= 2;
  // last sector of cluster has 2 bytes less
  return 0;  
}

void mmc_format()
{
  uint16_t temp16;
  uint8_t secondcnt = 0;
  led = 1;
  for(temp16 = 0; temp16 < 512; temp16++)
    buffer.b[temp16] = 0;
  for (temp16 = 0; temp16 <300; temp16++)
    mmc_write_sector(temp16, buffer.b);
  led = 2;
  uint32_t temp32;
  for(temp32 = FILETABLESIZE + CLUSTERSIZE; temp32 <= CLUSTERS * CLUSTERSIZE; temp32+=CLUSTERSIZE)
  {
    mmc_write_sector(temp32, buffer.b);
	secondcnt++;
	if(secondcnt == 0) led++;
  }
  /*for (clustercounter = FILETABLESIZE / CLUSTERSIZE; clustercounter <= CLUSTERS; clustercounter++)
  {
    mmc_write_sector(clustercounter * CLUSTERSIZE, buffer.b);	// first sector
	mmc_write_sector((clustercounter * CLUSTERSIZE) + (CLUSTERSIZE - 1), buffer.b); // last sector
	secondcnt++;
	if(secondcnt == 0) led++;
  }*/
}

void delete_file(uint8_t filenumber)
{
  for(uint16_t temp = 0; temp < 512; temp++)
    buffer.b[temp] = 0;
  
  uint8_t filetype = open_file(filenumber);
  mmc_read_sector_finish(0,1);
  mmc_write_sector(filenumber, buffer.b);
  if(filetype == FT_WAYPOINT)
	return;

  
  while(current_file.cluster.w)
  {
    mmc_write_sector(addr_mul_clustersize(current_file.cluster.w), buffer.b);
	mmc_read_sector_start(addr_mul_clustersize(current_file.cluster.w) + (CLUSTERSIZE - 1));
	mmc_read_sector_continue(0, 510 | 0x8000);		// discard bytes
	mmc_read_sector_continue(current_file.cluster.b, 2);
	mmc_read_sector_finish(0,1);
  }
}

void temp_swap_write_files()
{
  struct write_file_t temp;
  mmc_write_sector(addr_mul_clustersize(write_file.cluster.w) + write_file.sector_of_cluster, buffer.b);
  temp = paused_write_file;
  paused_write_file = write_file;
  write_file = temp;
  mmc_read_sector_start(addr_mul_clustersize(write_file.cluster.w) + write_file.sector_of_cluster);
  mmc_read_sector_finish(buffer.b,0);
}
