/*
  GPS Mini Navigator    for navigating predefined ways & extended gps logging

  Copyright (C) 2008  Matthias Larisch

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License along with
  this program; if not, see <http://www.gnu.org/licenses/>.
  

  file description: simple filesystem implementation
*/


#include "egl.h"
#include "filesystem_simple.h"
#include "mmc.h"

struct readfile_t
{
  uint8_t filenumber;
  uint8_t filetype;
  uint32_t sector;
  uint16_t bytes_remaining;
};

struct writefile_t
{
  uint8_t filetype;
  uint8_t filenumber;
  uint32_t sector;
  uint8_t *bufferw;
};

struct readfile_t readfile;
struct writefile_t writefile;
struct writefile_t temp_writefile;

union buffer_t buffer;
uint8_t glob_temp;					// helper variable, safes a lot when needing a pointer to an uint8_t

// Opens a file for reading, returns filetype
// does not finish SD transfer
uint8_t open_file(uint8_t filenumber)
{
  if(readfile.filetype != FT_NONE)
    close_read_file();
  mmc_read_sector_start(filenumber);
  mmc_read_sector_continue(&glob_temp, 1);
  readfile.filenumber = filenumber;
  readfile.filetype = glob_temp;
  readfile.bytes_remaining = BUFFER_SIZE - SECTORADDRESS_SIZE - 1;
  return glob_temp;
}

//reads bytes bytes from file to buffer
//seeks bytes forward if bytes & 0x8000
uint8_t read_file(uint8_t *buffer, uint16_t bytes)
{
  while(bytes >= readfile.bytes_remaining)				// this is not really needed/supported on Mega88
  														// because of low Sram, but added for compatibility
  {
    mmc_read_sector_continue(buffer, readfile.bytes_remaining);
	mmc_read_sector_finish((uint8_t *)&readfile.sector, 0);
	if(readfile.sector < FIRST_DATA_SECTOR || readfile.sector > LAST_DATA_SECTOR) return F_EOF;
	bytes -= readfile.bytes_remaining;
	readfile.bytes_remaining = BUFFER_SIZE - SECTORADDRESS_SIZE - 1;
	mmc_read_sector_start(readfile.sector);
	mmc_read_sector_continue(0,0x8001);
  }
  mmc_read_sector_continue(buffer, bytes);
  readfile.bytes_remaining -= bytes;
  return F_OK;
}

void close_read_file()
{
  mmc_read_sector_finish(0,1);
  readfile.filetype = 0;
}
//searches forward on the drive to find first unused sector (filetype entry = 0)
//speedup by saving last value. This causes not to use sectors of deleted files.
//to do so, device reset is needed. Warning: This can be somewhat slow on large
//drives.
uint32_t search_free_sector()
{
  static uint32_t search_offset = FIRST_DATA_SECTOR-1;
  glob_temp = 1;
  while(glob_temp)
  {
    mmc_read_sector_start(++search_offset);
	mmc_read_sector_continue(&glob_temp, 1);
	mmc_read_sector_finish(0,1);
	if(search_offset >= LAST_DATA_SECTOR) return F_MEDIUM_FULL;
  }
  return search_offset;
}

//creates file with type filetype, returns filenumber.
uint8_t create_file(uint8_t filetype)
{
  uint8_t i = 0;
  do
  {
    mmc_read_sector_start(i);
	mmc_read_sector_continue(&glob_temp,1);
	mmc_read_sector_finish(0,1);
	if(++i == 0) return F_NO_FREE_FILE_ENTRIES;
  } while (glob_temp);
  
  writefile.sector = (uint8_t)(i-1);
  //writefile.sector = search_free_sector();
  //if(writefile.sector == F_MEDIUM_FULL) return F_MEDIUM_FULL;
  buffer.b[0] = filetype;
  writefile.filetype = filetype;
  writefile.bufferw = &buffer.b[1];
  return writefile.sector;
}

//writes to currently opened writeable file, count bytes from buffer
uint8_t write_file(uint8_t *Buffer, uint8_t count)
{
  if(writefile.filetype == FT_NONE) return F_NO_FILE;
  while(writefile.bufferw < &buffer.b[BUFFER_SIZE - SECTORADDRESS_SIZE])
  //writes until sector is full or everything is written
  {
    *writefile.bufferw++ = *Buffer++;
	if(--count == 0) return F_OK;
  }
  buffer.q[BUFFER_SIZE/4 - 1] = search_free_sector();
  if(mmc_write_sector(writefile.sector, buffer.b)) return F_ERROR;
  buffer.b[0] = writefile.filetype;
  writefile.bufferw = &buffer.b[1];
  writefile.sector = buffer.q[BUFFER_SIZE/4 - 1];
  while(count--)
    *writefile.bufferw++ = *Buffer++;
  return F_NEW_SECTOR;
}

void delete_file(uint8_t filenumber)
{
  uint32_t next_sector;
  uint32_t this_sector = filenumber;
  uint16_t i = 0;
  while(i < 512) buffer.b[i++] = 0;
  do 
  {
    mmc_read_sector_start(this_sector);
    mmc_read_sector_continue(0, (SECTOR_SIZE - SECTORADDRESS_SIZE) | 0x8000);
    mmc_read_sector_finish((uint8_t *)&next_sector,0);
    mmc_write_sector(this_sector, buffer.b);
    this_sector = next_sector;
  } while(this_sector);
}

void close_write_file()
{
  while(writefile.bufferw < &buffer.b[BUFFER_SIZE])
    *writefile.bufferw++ = 0;
  mmc_write_sector(writefile.sector, buffer.b);
  writefile.filetype = 0;
}


// makes two simultaneous writes possible through file handle buffer. Warning: On every call,
// a sector gets written! that may wear out card lifetime faster than usual.
void swap_write_files()
{
  if(writefile.filetype != 0)
  {
    mmc_write_sector(writefile.sector, buffer.b);
  }
  struct writefile_t tempfile;
  tempfile = temp_writefile;
  temp_writefile = writefile;
  writefile = tempfile;
  if(writefile.filetype != 0)
  {
    mmc_read_sector_start(writefile.sector);
	mmc_read_sector_finish(buffer.b, 0);
  }
}
