/************************************************************************/
/*                                                                      */
/*                      Software UART using T1                          */
/*                                                                      */
/*              Author: P. Dannegger                                    */
/*                      danni@specs.de                                  */
/*                                                                      */
/************************************************************************/
#include "egl.h"
#include "suart.h"


#define BIT_TIME	(uint16_t)((XTAL + SW_BAUD/2) / SW_BAUD)


volatile uint8_t stx_count;
uint8_t stx_data;

volatile uint8_t srx_done;
uint8_t srx_data;
uint8_t srx_mask;
uint8_t srx_tmp;


void suart_init( void )
{
  OCR1A = TCNT1 + 1;			// force first compare
  TCCR1A = 1<<COM1A1^1<<COM1A0;		// set OC1A high, T1 mode 0
  TCCR1B = 1<<ICNC1^1<<CS10;		// noise canceler, 1>0 transition,
					// CLK/1, T1 mode 0
  TIFR1 = 1<<ICF1;			// clear pending interrupt
  TIMSK1 = 1<<ICIE1^1<<OCIE1A;		// enable tx and wait for start

  stx_count = 0;			// nothing to sent
  srx_done = 0;				// nothing received
  power_status.sw_uart_on = 1;
  
}

void suart_deinit( void )
{
  TCCR1A = 0;
  TCCR1B = 0;
  TIMSK1 = 0;
  power_status.sw_uart_on = 0;
}


uint8_t sgetchar( void )			// get byte
{
  while( !srx_done );			// wait until byte received
  srx_done = 0;
  return srx_data;
}


ISR(TIMER1_CAPT_vect)		// rx start  (55 clocks including call/reti)
{
  OCR1B = ICR1 + (uint16_t)(BIT_TIME*1.2);	// scan 1.2 bits after start
  // this was changed by Matthias Larisch. (original: 1.5 * BIT_TIME)
  // Idea: @ 8 MHz and 38400 baud we have 208 cycles per bit.
  // when we trigger our receive interrupt at 1.2 * BIT_TIME, we can be
  // max. 167 cycles late, else we will miss a bit (correct: we wont miss it
  // but we will get a wrong value. For a miss, we must be late a lot more)
  // Because of the interrupt call, we are already 19-21 cycles late everytime.
  // that means we have 148 cycles MAX for everything else that may occur
  // during suart receives. 48 cycles for the transmit interrupt with empty
  // transmit byte. when both interrupts occur at the same time, TX int has
  // a higher priority, so we have only 97-99 cycles left. The Timer 1 OVF int
  // needs about 100-110 cycles. we can only lower BIT_TIME factor more...

  srx_tmp = 0;				// clear bit storage
  srx_mask = 1;				// bit mask
  TIFR1 = 1<<OCF1B;			// clear pending interrupt
  if( !(SRXPIN & 1<<SRX))		// still low
    TIMSK1 = 1<<OCIE1A^1<<OCIE1B;	// wait for first bit
}


ISR(TIMER1_COMPB_vect)			// (~61 clocks including call/reti)
{
  uint8_t in = SRXPIN;			// scan rx line

  if( srx_mask ){
    if( in & 1<<SRX )
      srx_tmp |= srx_mask;
    srx_mask <<= 1;
    OCR1B += BIT_TIME;			// next bit slice
  }else{
    srx_done = 1;			// mark rx data valid
    srx_data = srx_tmp;			// store rx data
    TIFR1 = 1<<ICF1;			// clear pending interrupt
    TIMSK1 = 1<<ICIE1^1<<OCIE1A;        // enable tx and wait for start
	bluetooth_on_cnt = 0;			// reset bluetooth shutdown counter
  }
}


void sputchar( uint8_t val )			// send byte
{
  while( stx_count );			// until last byte finished
  stx_data = ~val;			// invert data for Stop bit generation
  stx_count = 10;			// 10 bits: Start + data + Stop
}


void sputs( uint8_t *txt )			// send string
{
  while( *txt )
    sputchar( *txt++ );
}


ISR(TIMER1_COMPA_vect)		// tx bit (65 clocks including call/reti)
{
  uint8_t dout;
  uint8_t count;

  OCR1A += BIT_TIME;			// next bit slice
  count = stx_count;

  if( count ){
    stx_count = --count;		// count down
    dout = 1<<COM1A1;			// set low on next compare
    if( count != 9 ){			// no start bit
      if( !(stx_data & 1) )		// test inverted data
	dout = 1<<COM1A1^1<<COM1A0;	// set high on next compare
      stx_data >>= 1;			// shift zero in from left
    }
    TCCR1A = dout;
  }
}
